# from keras_unet_collection.models import unet_3plus_2d
from keras_unet_collection.models import unet_3plus_2d
from tqdm import tqdm
input_size = (128, 128, 3)
n_labels = 3
filter_num_down = [32, 64, 128, 256, 512]
# filter_num_skip = [32, 32, 32, 32]
# filter_num_aggregate = 160
filter_num_skip='auto'
filter_num_aggregate='auto'

stack_num_down=2
stack_num_up=1
activation='ReLU'
output_activation='Sigmoid',
batch_norm=False
pool=True
unpool=True
deep_supervision=False

backbone=None
weights='imagenet'
freeze_backbone=True
freeze_batch_norm=True
name='unet3plus'
                  
model = unet_3plus_2d(input_size, n_labels, filter_num_down)

from keras_unet_collection import losses

def hybrid_loss(y_true, y_pred):
    print('y_true: ', y_true.shape)
    print('y_pred: ', y_pred.shape)

    loss_focal = losses.focal_tversky(y_true, y_pred, alpha=0.5, gamma=4/3)
    loss_iou = losses.iou_seg(y_true, y_pred)
    
    # (x) 
    #loss_ssim = losses.ms_ssim(y_true, y_pred, max_val=1.0, filter_size=4)
    
    return loss_focal+loss_iou #+loss_ssim

from glob import glob
import numpy as np
import os
os.environ["CUDA_VISIBLE_DEVICES"] = "0,1"  

import tensorflow as tf 
print("Num GPUs Available: ", len(tf.config.list_physical_devices('GPU')))
gpus = tf.config.list_physical_devices('GPU')
if gpus:
    try:
        # 메모리 증가 허용
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.list_logical_devices('GPU')
        print(f"{len(gpus)} Physical GPUs, {len(logical_gpus)} Logical GPUs")
    except RuntimeError as e:
        print(e)
from tensorflow import keras
import keras_unet_collection.utils as utils

def input_data_process(input_array):
    '''converting pixel vales to [0, 1]'''
    return input_array/255.

def target_data_process(target_array):
    '''Converting tri-mask of {1, 2, 3} to three categories.'''
    return keras.utils.to_categorical(target_array-1)

# input_dir = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/patches'
# mask_input_dir = '/HDD/_projects/benchmark/semantic_segmentation/new_model/datasets/masks'

# sample_names = np.array(sorted(glob(input_dir + '/*.bmp')))
# label_names = np.array(sorted(glob(mask_input_dir + '/*.bmp')))

input_dir = '/HDD/datasets/public/Oxford_IIIT/images'
mask_input_dir = '/HDD/datasets/public/Oxford_IIIT/annotations/trimaps'

sample_names = np.array(sorted(glob(input_dir + '/*.jpg')))
label_names = np.array(sorted(glob(mask_input_dir + '/*.png')))

L = len(sample_names)
ind_all = utils.shuffle_ind(L)

L_train = int(0.1*L)
L_valid = int(0.1*L)
L_test = int(0.1*L)
ind_train = ind_all[:L_train]
ind_valid = ind_all[L_train:L_train+L_valid]
ind_test = ind_all[L_train+L_valid:]
print("Training:validation:testing = {}:{}:{}".format(L_train, L_valid, L_test))

valid_input = input_data_process(utils.image_to_array(sample_names[ind_valid], size=128, channel=3))
valid_target = target_data_process(utils.image_to_array(label_names[ind_valid], size=128, channel=1))

# test_input = input_data_process(utils.image_to_array(sample_names[ind_test], size=128, channel=3))
# test_target = target_data_process(utils.image_to_array(label_names[ind_test], size=128, channel=1))

N_epoch = 10 # number of epoches
N_batch = 100 # number of batches per epoch
N_sample = 2 # number of samples per batch

tol = 0 # current early stopping patience
max_tol = 3 # the max-allowed early stopping patience
min_del = 0 # the lowest acceptable loss value reduction 

model.compile(loss=[hybrid_loss, hybrid_loss, hybrid_loss, hybrid_loss, hybrid_loss],
                  loss_weights=[0.25, 0.25, 0.25, 0.25, 1.0],
                  optimizer=keras.optimizers.Adam(learning_rate=1e-4))

# loop over epoches
for epoch in range(N_epoch):
    if epoch%30 == 0:
        model.save(f'/HDD/unet3p_{epoch}.h5')
    
    # loop over batches
    for step in tqdm(range(N_batch), desc=epoch):
        # selecting smaples for the current batch
        ind_train_shuffle = utils.shuffle_ind(L_train)[:N_sample]
        
        train_input = input_data_process(
            utils.image_to_array(sample_names[ind_train][ind_train_shuffle], size=128, channel=3))
        train_target = target_data_process(
            utils.image_to_array(label_names[ind_train][ind_train_shuffle], size=128, channel=1))
        
        loss_ = model.train_on_batch([train_input,], 
                                         [train_target, train_target, train_target, train_target, train_target,])
        if np.isnan(loss_):
            print("Training blow-up")



import torch
import torch.nn as nn
import torch.nn.functional as F
from .init_weights import init_weights


class unetConv2(nn.Module):
    def __init__(self, in_size, out_size, is_batchnorm, n=2, ks=3, stride=1, padding=1):
        super(unetConv2, self).__init__()
        self.n = n
        self.ks = ks
        self.stride = stride
        self.padding = padding
        s = stride
        p = padding
        
        if is_batchnorm:
            for i in range(1, n + 1):
                conv = nn.Sequential(nn.Conv2d(in_size, out_size, ks, s, p),
                                     nn.BatchNorm2d(out_size), nn.ReLU(inplace=True),)
                setattr(self, 'conv%d' % i, conv)
                in_size = out_size
        else:
            for i in range(1, n + 1):
                conv = nn.Sequential(nn.Conv2d(in_size, out_size, ks, s, p), nn.ReLU(inplace=True), )
                setattr(self, 'conv%d' % i, conv)
                in_size = out_size

        # initialise the blocks
        for m in self.children():
            init_weights(m, init_type='kaiming')


    def forward(self, inputs):
        x = inputs
        for i in range(1, self.n + 1):
            conv = getattr(self, 'conv%d' % i)
            x = conv(x)
        return x


class unetUp(nn.Module):
    def __init__(self, in_size, out_size, is_deconv, n_concat=2):
        super(unetUp, self).__init__()
        self.conv = unetConv2(out_size * 2, out_size, False)
        
        if is_deconv:
            self.up = nn.ConvTranspose2d(in_size, out_size, kernel_size=4, stride=2, padding=1)
        else:
            self.up = nn.UpsamplingBilinear2d(scale_factor=2)

        # initialise the blocks
        for m in self.children():
            if m.__class__.__name__.find('unetConv2') != -1: continue
            init_weights(m, init_type='kaiming')


    def forward(self, inputs0, *input):
        outputs0 = self.up(inputs0)
        for i in range(len(input)):
            outputs0 = torch.cat([outputs0, input[i]], 1)
        return self.conv(outputs0)

    
class unetUp_origin(nn.Module):
    def __init__(self, in_size, out_size, is_deconv, n_concat=2):
        super(unetUp_origin, self).__init__()
        # self.conv = unetConv2(out_size*2, out_size, False)
        if is_deconv:
            self.conv = unetConv2(in_size + (n_concat - 2) * out_size, out_size, False)
            self.up = nn.ConvTranspose2d(in_size, out_size, kernel_size=4, stride=2, padding=1)
        else:
            self.conv = unetConv2(in_size + (n_concat - 2) * out_size, out_size, False)
            self.up = nn.UpsamplingBilinear2d(scale_factor=2)

        # initialise the blocks
        for m in self.children():
            if m.__class__.__name__.find('unetConv2') != -1: continue
            init_weights(m, init_type='kaiming')

    def forward(self, inputs0, *input):
        # print(self.n_concat)
        # print(input)
        outputs0 = self.up(inputs0)
        for i in range(len(input)):
            outputs0 = torch.cat([outputs0, input[i]], 1)
        return self.conv(outputs0)



import tensorflow as tf
import numpy as np
from scipy.ndimage import distance_transform_edt as distance
from label_smooth import LabelSmoothSoftmaxCEV1
from functools import partial
from operator import itemgetter

# Helper functions
def kl_div(a, b):
    return tf.nn.softmax(b, axis=1) * (tf.nn.log_softmax(b, axis=1) - tf.nn.log_softmax(a, axis=1))

def one_hot2dist(seg):
    seg = tf.convert_to_tensor(seg)
    res = tf.zeros_like(seg, dtype=tf.float32)
    
    for i in range(tf.shape(seg)[0]):
        posmask = tf.cast(seg[i], tf.bool)
        if tf.reduce_any(posmask):
            negmask = tf.logical_not(posmask)
            res_i = tf.where(negmask, tf.cast(distance(negmask.numpy()), tf.float32), 0.0)
            res_i = res_i - tf.where(posmask, tf.cast(distance(posmask.numpy()) - 1, tf.float32), 0.0)
            res = tf.tensor_scatter_nd_update(res, [[i]], [res_i])
    
    return res


def class2one_hot(seg, C):
    seg = tf.expand_dims(seg, axis=0) if len(seg.shape) == 2 else seg
    res = tf.stack([tf.cast(seg == c, tf.int32) for c in range(C)], axis=1)
    return res

class ABL(tf.keras.layers.Layer):
    def __init__(self, isdetach=True, max_N_ratio=1/100, ignore_label=255, label_smoothing=0.2, weight=None, max_clip_dist=20.):
        super(ABL, self).__init__()
        self.ignore_label = ignore_label
        self.label_smoothing = label_smoothing
        self.isdetach = isdetach
        self.max_N_ratio = max_N_ratio
        self.weight_func = lambda w, max_distance=max_clip_dist: tf.clip_by_value(w, 0, max_distance) / max_distance

        self.dist_map_transform = tf.keras.Sequential([
            tf.keras.layers.Lambda(lambda x: tf.expand_dims(x, axis=0)),
            tf.keras.layers.Lambda(lambda x: tf.cast(x, tf.int64)),
            tf.keras.layers.Lambda(partial(class2one_hot, C=1)),
            tf.keras.layers.Lambda(lambda x: x[0]),  # Replace itemgetter(0) with this
            tf.keras.layers.Lambda(lambda x: tf.py_function(one_hot2dist, [x], tf.float32))
        ])

        if label_smoothing == 0:
            self.criterion = tf.keras.losses.SparseCategoricalCrossentropy(
                from_logits=True, reduction=tf.keras.losses.Reduction.NONE
            )
        else:
            self.criterion = LabelSmoothSoftmaxCEV1(
                reduction='none', ignore_index=ignore_label, lb_smooth=label_smoothing
            )

    def logits2boundary(self, logit):
        eps = 1e-5
        _, h, w, _ = logit.shape
        max_N = tf.cast(h * w * self.max_N_ratio, tf.float32)
        kl_ud = tf.reduce_sum(kl_div(logit[:, 1:, :, :], logit[:, :-1, :, :]), axis=-1, keepdims=True)
        kl_lr = tf.reduce_sum(kl_div(logit[:, :, 1:, :], logit[:, :, :-1, :]), axis=-1, keepdims=True)
        kl_ud = tf.pad(kl_ud, [[0, 0], [0, 1], [0, 0], [0, 0]], mode='CONSTANT', constant_values=0)
        kl_lr = tf.pad(kl_lr, [[0, 0], [0, 0], [0, 1], [0, 0]], mode='CONSTANT', constant_values=0)
        kl_combine = kl_lr + kl_ud

        while True:
            kl_combine_bin = tf.cast(kl_combine > eps, tf.float32)
            if tf.reduce_sum(kl_combine_bin) > max_N:
                eps *= 1.2
            else:
                break

        dilate_weight = tf.ones((3, 3, 1, 1))
        edge2 = tf.nn.conv2d(kl_combine_bin, dilate_weight, strides=[1, 1, 1, 1], padding='SAME')
        edge2 = tf.squeeze(edge2, axis=-1)
        kl_combine_bin = edge2 > 0
        return kl_combine_bin

    def gt2boundary(self, gt, ignore_label=-1):
        gt_ud = gt[:, 1:, :] - gt[:, :-1, :]
        gt_lr = gt[:, :, 1:] - gt[:, :, :-1]
        gt_ud = tf.pad(gt_ud, [[0, 0], [0, 1], [0, 0]]) != 0
        gt_lr = tf.pad(gt_lr, [[0, 0], [0, 0], [0, 1]]) != 0
        gt_combine = gt_lr | gt_ud
        gt_combine = gt_combine | tf.equal(gt, ignore_label)
        return gt_combine

    def get_direction_gt_predkl(self, pred_dist_map, pred_bound, logits):
        eps = 1e-5
        bound = tf.where(pred_bound)
        n, x, y = tf.unstack(bound, axis=1)
        max_dis = 1e5

        logits = tf.transpose(logits, [0, 2, 3, 1])
        pred_dist_map_d = tf.pad(pred_dist_map, [[0, 0], [1, 1], [1, 1]], mode='CONSTANT', constant_values=max_dis)
        logits_d = tf.pad(logits, [[0, 0], [1, 1], [1, 1], [0, 0]], mode='CONSTANT')

        x_range = [1, -1, 0, 0, -1, 1, -1, 1, 0]
        y_range = [0, 0, -1, 1, 1, 1, -1, -1, 0]
        dist_maps = []
        kl_maps = []

        kl_center = tf.gather_nd(logits, tf.stack([n, x, y], axis=1))

        for dx, dy in zip(x_range, y_range):
            dist_now = tf.gather_nd(pred_dist_map_d, tf.stack([n, x+dx+1, y+dy+1], axis=1))
            dist_maps.append(dist_now)

            if dx != 0 or dy != 0:
                logits_now = tf.gather_nd(logits_d, tf.stack([n, x+dx+1, y+dy+1], axis=1))
                if self.isdetach:
                    logits_now = tf.stop_gradient(logits_now)
                kl_map_now = kl_div(kl_center, logits_now)
                kl_map_now = tf.reduce_sum(kl_map_now, axis=1)
                kl_maps.append(kl_map_now)

        dist_maps = tf.stack(dist_maps, axis=0)
        kl_maps = tf.stack(kl_maps[:-1], axis=0)  # Exclude the last element (center)

        direction_gt = tf.argmin(dist_maps, axis=0)
        weight_ce = tf.gather_nd(pred_dist_map, tf.stack([n, x, y], axis=1))

        direction_gt_idx = direction_gt != 8
        direction_gt = tf.boolean_mask(direction_gt, direction_gt_idx)
        kl_maps = tf.transpose(kl_maps, [1, 0])
        direction_pred = tf.boolean_mask(kl_maps, direction_gt_idx)
        weight_ce = tf.boolean_mask(weight_ce, direction_gt_idx)

        return direction_gt, direction_pred, weight_ce

    def get_dist_maps(self, target):
        target_detach = tf.identity(target)
        dist_maps = tf.concat([self.dist_map_transform(target_detach[i]) for i in range(target_detach.shape[0])], axis=0)
        out = -dist_maps
        out = tf.where(out > 0, out, tf.zeros_like(out))
        return out

    def call(self, logits, target):
        eps = 1e-10
        ph, pw = tf.shape(logits)[2], tf.shape(logits)[3]
        h, w = tf.shape(target)[1], tf.shape(target)[2]

        if ph != h or pw != w:
            logits = tf.image.resize(logits, [h, w], method='bilinear')

        gt_boundary = self.gt2boundary(target, ignore_label=self.ignore_label)
        dist_maps = self.get_dist_maps(gt_boundary)
        pred_boundary = self.logits2boundary(logits)

        if tf.reduce_sum(tf.cast(pred_boundary, tf.float32)) < 1:
            return None

        direction_gt, direction_pred, weight_ce = self.get_direction_gt_predkl(dist_maps, pred_boundary, logits)
        loss = self.criterion(direction_pred, direction_gt)
        weight_ce = self.weight_func(weight_ce)
        loss = tf.reduce_mean(loss * weight_ce)

        return loss
    
if __name__ == '__main__':

    import tensorflow as tf
    import numpy as np
    import random
    import os

    seed = 0
    np.random.seed(seed)
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)


    # Define test dimensions
    n, c, h, w = 1, 2, 100, 100

    # Create ground truth tensor
    gt = np.zeros((n, h, w))
    gt = np.zeros_like(gt)  # Initialize with zeros if not already done

    # Perform the scatter update
    gt[0, 5, 0] = 1.0
    gt[0, 50, 0] = 1.0

    logits = np.random.normal(loc=0.0, scale=4.0, size=(n, c, h, w)).astype(np.float32)
    tf_logits = tf.constant(logits)
    tf_gt = tf.constant(gt)
    # tf.config.set_visible_devices([], 'GPU')
    tf_abl = ABL()
    tf_loss = tf_abl(tf_logits, tf_gt)

    print("TensorFlow ABL Loss:", tf_loss.numpy())
    
    from visionsuite.utils.losses.torch.active_boundary_loss import ABL as TorchABL
    import torch 
    
    torch_logits = torch.from_numpy(logits)
    torch_gt = torch.from_numpy(gt).to(torch.int64)
    torch_abl = TorchABL()
    torch_loss = torch_abl(torch_logits, torch_gt)
    print("Torch ABL Loss:", tf_loss.numpy())
    
    
    



# Copyright (c) Open-MMLab. All rights reserved.
import io
import os
import os.path as osp
import pkgutil
import time
import warnings
from collections import OrderedDict
from importlib import import_module
from tempfile import TemporaryDirectory

import torch
import torchvision
from torch.optim import Optimizer
from torch.utils import model_zoo
from torch.nn import functional as F

import mmcv
from mmcv.fileio import FileClient
from mmcv.fileio import load as load_file
from mmcv.parallel import is_module_wrapper
from mmcv.utils import mkdir_or_exist
from mmcv.runner import get_dist_info

ENV_MMCV_HOME = 'MMCV_HOME'
ENV_XDG_CACHE_HOME = 'XDG_CACHE_HOME'
DEFAULT_CACHE_DIR = '~/.cache'


def _get_mmcv_home():
    mmcv_home = os.path.expanduser(
        os.getenv(
            ENV_MMCV_HOME,
            os.path.join(
                os.getenv(ENV_XDG_CACHE_HOME, DEFAULT_CACHE_DIR), 'mmcv')))

    mkdir_or_exist(mmcv_home)
    return mmcv_home


def load_state_dict(module, state_dict, strict=False, logger=None):
    """Load state_dict to a module.

    This method is modified from :meth:`torch.nn.Module.load_state_dict`.
    Default value for ``strict`` is set to ``False`` and the message for
    param mismatch will be shown even if strict is False.

    Args:
        module (Module): Module that receives the state_dict.
        state_dict (OrderedDict): Weights.
        strict (bool): whether to strictly enforce that the keys
            in :attr:`state_dict` match the keys returned by this module's
            :meth:`~torch.nn.Module.state_dict` function. Default: ``False``.
        logger (:obj:`logging.Logger`, optional): Logger to log the error
            message. If not specified, print function will be used.
    """
    unexpected_keys = []
    all_missing_keys = []
    err_msg = []

    metadata = getattr(state_dict, '_metadata', None)
    state_dict = state_dict.copy()
    if metadata is not None:
        state_dict._metadata = metadata

    # use _load_from_state_dict to enable checkpoint version control
    def load(module, prefix=''):
        # recursively check parallel module in case that the model has a
        # complicated structure, e.g., nn.Module(nn.Module(DDP))
        if is_module_wrapper(module):
            module = module.module
        local_metadata = {} if metadata is None else metadata.get(
            prefix[:-1], {})
        module._load_from_state_dict(state_dict, prefix, local_metadata, True,
                                     all_missing_keys, unexpected_keys,
                                     err_msg)
        for name, child in module._modules.items():
            if child is not None:
                load(child, prefix + name + '.')

    load(module)
    load = None  # break load->load reference cycle

    # ignore "num_batches_tracked" of BN layers
    missing_keys = [
        key for key in all_missing_keys if 'num_batches_tracked' not in key
    ]

    if unexpected_keys:
        err_msg.append('unexpected key in source '
                       f'state_dict: {", ".join(unexpected_keys)}\n')
    if missing_keys:
        err_msg.append(
            f'missing keys in source state_dict: {", ".join(missing_keys)}\n')

    rank, _ = get_dist_info()
    if len(err_msg) > 0 and rank == 0:
        err_msg.insert(
            0, 'The model and loaded state dict do not match exactly\n')
        err_msg = '\n'.join(err_msg)
        if strict:
            raise RuntimeError(err_msg)
        elif logger is not None:
            logger.warning(err_msg)
        else:
            print(err_msg)


def load_url_dist(url, model_dir=None):
    """In distributed setting, this function only download checkpoint at local
    rank 0."""
    rank, world_size = get_dist_info()
    rank = int(os.environ.get('LOCAL_RANK', rank))
    if rank == 0:
        checkpoint = model_zoo.load_url(url, model_dir=model_dir)
    if world_size > 1:
        torch.distributed.barrier()
        if rank > 0:
            checkpoint = model_zoo.load_url(url, model_dir=model_dir)
    return checkpoint


def load_pavimodel_dist(model_path, map_location=None):
    """In distributed setting, this function only download checkpoint at local
    rank 0."""
    try:
        from pavi import modelcloud
    except ImportError:
        raise ImportError(
            'Please install pavi to load checkpoint from modelcloud.')
    rank, world_size = get_dist_info()
    rank = int(os.environ.get('LOCAL_RANK', rank))
    if rank == 0:
        model = modelcloud.get(model_path)
        with TemporaryDirectory() as tmp_dir:
            downloaded_file = osp.join(tmp_dir, model.name)
            model.download(downloaded_file)
            checkpoint = torch.load(downloaded_file, map_location=map_location)
    if world_size > 1:
        torch.distributed.barrier()
        if rank > 0:
            model = modelcloud.get(model_path)
            with TemporaryDirectory() as tmp_dir:
                downloaded_file = osp.join(tmp_dir, model.name)
                model.download(downloaded_file)
                checkpoint = torch.load(
                    downloaded_file, map_location=map_location)
    return checkpoint


def load_fileclient_dist(filename, backend, map_location):
    """In distributed setting, this function only download checkpoint at local
    rank 0."""
    rank, world_size = get_dist_info()
    rank = int(os.environ.get('LOCAL_RANK', rank))
    allowed_backends = ['ceph']
    if backend not in allowed_backends:
        raise ValueError(f'Load from Backend {backend} is not supported.')
    if rank == 0:
        fileclient = FileClient(backend=backend)
        buffer = io.BytesIO(fileclient.get(filename))
        checkpoint = torch.load(buffer, map_location=map_location)
    if world_size > 1:
        torch.distributed.barrier()
        if rank > 0:
            fileclient = FileClient(backend=backend)
            buffer = io.BytesIO(fileclient.get(filename))
            checkpoint = torch.load(buffer, map_location=map_location)
    return checkpoint


def get_torchvision_models():
    model_urls = dict()
    for _, name, ispkg in pkgutil.walk_packages(torchvision.models.__path__):
        if ispkg:
            continue
        _zoo = import_module(f'torchvision.models.{name}')
        if hasattr(_zoo, 'model_urls'):
            _urls = getattr(_zoo, 'model_urls')
            model_urls.update(_urls)
    return model_urls


def get_external_models():
    mmcv_home = _get_mmcv_home()
    default_json_path = osp.join(mmcv.__path__[0], 'model_zoo/open_mmlab.json')
    default_urls = load_file(default_json_path)
    assert isinstance(default_urls, dict)
    external_json_path = osp.join(mmcv_home, 'open_mmlab.json')
    if osp.exists(external_json_path):
        external_urls = load_file(external_json_path)
        assert isinstance(external_urls, dict)
        default_urls.update(external_urls)

    return default_urls


def get_mmcls_models():
    mmcls_json_path = osp.join(mmcv.__path__[0], 'model_zoo/mmcls.json')
    mmcls_urls = load_file(mmcls_json_path)

    return mmcls_urls


def get_deprecated_model_names():
    deprecate_json_path = osp.join(mmcv.__path__[0],
                                   'model_zoo/deprecated.json')
    deprecate_urls = load_file(deprecate_json_path)
    assert isinstance(deprecate_urls, dict)

    return deprecate_urls


def _process_mmcls_checkpoint(checkpoint):
    state_dict = checkpoint['state_dict']
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        if k.startswith('backbone.'):
            new_state_dict[k[9:]] = v
    new_checkpoint = dict(state_dict=new_state_dict)

    return new_checkpoint


def _load_checkpoint(filename, map_location=None):
    """Load checkpoint from somewhere (modelzoo, file, url).

    Args:
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``. Please refer to ``docs/model_zoo.md`` for
            details.
        map_location (str | None): Same as :func:`torch.load`. Default: None.

    Returns:
        dict | OrderedDict: The loaded checkpoint. It can be either an
            OrderedDict storing model weights or a dict containing other
            information, which depends on the checkpoint.
    """
    if filename.startswith('modelzoo://'):
        warnings.warn('The URL scheme of "modelzoo://" is deprecated, please '
                      'use "torchvision://" instead')
        model_urls = get_torchvision_models()
        model_name = filename[11:]
        checkpoint = load_url_dist(model_urls[model_name])
    elif filename.startswith('torchvision://'):
        model_urls = get_torchvision_models()
        model_name = filename[14:]
        checkpoint = load_url_dist(model_urls[model_name])
    elif filename.startswith('open-mmlab://'):
        model_urls = get_external_models()
        model_name = filename[13:]
        deprecated_urls = get_deprecated_model_names()
        if model_name in deprecated_urls:
            warnings.warn(f'open-mmlab://{model_name} is deprecated in favor '
                          f'of open-mmlab://{deprecated_urls[model_name]}')
            model_name = deprecated_urls[model_name]
        model_url = model_urls[model_name]
        # check if is url
        if model_url.startswith(('http://', 'https://')):
            checkpoint = load_url_dist(model_url)
        else:
            filename = osp.join(_get_mmcv_home(), model_url)
            if not osp.isfile(filename):
                raise IOError(f'{filename} is not a checkpoint file')
            checkpoint = torch.load(filename, map_location=map_location)
    elif filename.startswith('mmcls://'):
        model_urls = get_mmcls_models()
        model_name = filename[8:]
        checkpoint = load_url_dist(model_urls[model_name])
        checkpoint = _process_mmcls_checkpoint(checkpoint)
    elif filename.startswith(('http://', 'https://')):
        checkpoint = load_url_dist(filename)
    elif filename.startswith('pavi://'):
        model_path = filename[7:]
        checkpoint = load_pavimodel_dist(model_path, map_location=map_location)
    elif filename.startswith('s3://'):
        checkpoint = load_fileclient_dist(
            filename, backend='ceph', map_location=map_location)
    else:
        if not osp.isfile(filename):
            raise IOError(f'{filename} is not a checkpoint file')
        checkpoint = torch.load(filename, map_location=map_location)
    return checkpoint


def load_checkpoint(model,
                    filename,
                    map_location='cpu',
                    strict=False,
                    logger=None):
    """Load checkpoint from a file or URI.

    Args:
        model (Module): Module to load checkpoint.
        filename (str): Accept local filepath, URL, ``torchvision://xxx``,
            ``open-mmlab://xxx``. Please refer to ``docs/model_zoo.md`` for
            details.
        map_location (str): Same as :func:`torch.load`.
        strict (bool): Whether to allow different params for the model and
            checkpoint.
        logger (:mod:`logging.Logger` or None): The logger for error message.

    Returns:
        dict or OrderedDict: The loaded checkpoint.
    """
    checkpoint = _load_checkpoint(filename, map_location)
    # OrderedDict is a subclass of dict
    if not isinstance(checkpoint, dict):
        raise RuntimeError(
            f'No state_dict found in checkpoint file {filename}')
    # get state_dict from checkpoint
    if 'state_dict' in checkpoint:
        state_dict = checkpoint['state_dict']
    elif 'model' in checkpoint:
        state_dict = checkpoint['model']
    else:
        state_dict = checkpoint
    # strip prefix of state_dict
    if list(state_dict.keys())[0].startswith('module.'):
        state_dict = {k[7:]: v for k, v in state_dict.items()}

    # for MoBY, load model of online branch
    if sorted(list(state_dict.keys()))[0].startswith('encoder'):
        state_dict = {k.replace('encoder.', ''): v for k, v in state_dict.items() if k.startswith('encoder.')}

    # reshape absolute position embedding
    if state_dict.get('absolute_pos_embed') is not None:
        absolute_pos_embed = state_dict['absolute_pos_embed']
        N1, L, C1 = absolute_pos_embed.size()
        N2, C2, H, W = model.absolute_pos_embed.size()
        if N1 != N2 or C1 != C2 or L != H*W:
            logger.warning("Error in loading absolute_pos_embed, pass")
        else:
            state_dict['absolute_pos_embed'] = absolute_pos_embed.view(N2, H, W, C2).permute(0, 3, 1, 2)

    # interpolate position bias table if needed
    relative_position_bias_table_keys = [k for k in state_dict.keys() if "relative_position_bias_table" in k]
    for table_key in relative_position_bias_table_keys:
        table_pretrained = state_dict[table_key]
        table_current = model.state_dict()[table_key]
        L1, nH1 = table_pretrained.size()
        L2, nH2 = table_current.size()
        if nH1 != nH2:
            logger.warning(f"Error in loading {table_key}, pass")
        else:
            if L1 != L2:
                S1 = int(L1 ** 0.5)
                S2 = int(L2 ** 0.5)
                table_pretrained_resized = F.interpolate(
                     table_pretrained.permute(1, 0).view(1, nH1, S1, S1),
                     size=(S2, S2), mode='bicubic')
                state_dict[table_key] = table_pretrained_resized.view(nH2, L2).permute(1, 0)

    # load state_dict
    load_state_dict(model, state_dict, strict, logger)
    return checkpoint


def weights_to_cpu(state_dict):
    """Copy a model state_dict to cpu.

    Args:
        state_dict (OrderedDict): Model weights on GPU.

    Returns:
        OrderedDict: Model weights on GPU.
    """
    state_dict_cpu = OrderedDict()
    for key, val in state_dict.items():
        state_dict_cpu[key] = val.cpu()
    return state_dict_cpu


def _save_to_state_dict(module, destination, prefix, keep_vars):
    """Saves module state to `destination` dictionary.

    This method is modified from :meth:`torch.nn.Module._save_to_state_dict`.

    Args:
        module (nn.Module): The module to generate state_dict.
        destination (dict): A dict where state will be stored.
        prefix (str): The prefix for parameters and buffers used in this
            module.
    """
    for name, param in module._parameters.items():
        if param is not None:
            destination[prefix + name] = param if keep_vars else param.detach()
    for name, buf in module._buffers.items():
        # remove check of _non_persistent_buffers_set to allow nn.BatchNorm2d
        if buf is not None:
            destination[prefix + name] = buf if keep_vars else buf.detach()


def get_state_dict(module, destination=None, prefix='', keep_vars=False):
    """Returns a dictionary containing a whole state of the module.

    Both parameters and persistent buffers (e.g. running averages) are
    included. Keys are corresponding parameter and buffer names.

    This method is modified from :meth:`torch.nn.Module.state_dict` to
    recursively check parallel module in case that the model has a complicated
    structure, e.g., nn.Module(nn.Module(DDP)).

    Args:
        module (nn.Module): The module to generate state_dict.
        destination (OrderedDict): Returned dict for the state of the
            module.
        prefix (str): Prefix of the key.
        keep_vars (bool): Whether to keep the variable property of the
            parameters. Default: False.

    Returns:
        dict: A dictionary containing a whole state of the module.
    """
    # recursively check parallel module in case that the model has a
    # complicated structure, e.g., nn.Module(nn.Module(DDP))
    if is_module_wrapper(module):
        module = module.module

    # below is the same as torch.nn.Module.state_dict()
    if destination is None:
        destination = OrderedDict()
        destination._metadata = OrderedDict()
    destination._metadata[prefix[:-1]] = local_metadata = dict(
        version=module._version)
    _save_to_state_dict(module, destination, prefix, keep_vars)
    for name, child in module._modules.items():
        if child is not None:
            get_state_dict(
                child, destination, prefix + name + '.', keep_vars=keep_vars)
    for hook in module._state_dict_hooks.values():
        hook_result = hook(module, destination, prefix, local_metadata)
        if hook_result is not None:
            destination = hook_result
    return destination


def save_checkpoint(model, filename, optimizer=None, meta=None):
    """Save checkpoint to file.

    The checkpoint will have 3 fields: ``meta``, ``state_dict`` and
    ``optimizer``. By default ``meta`` will contain version and time info.

    Args:
        model (Module): Module whose params are to be saved.
        filename (str): Checkpoint filename.
        optimizer (:obj:`Optimizer`, optional): Optimizer to be saved.
        meta (dict, optional): Metadata to be saved in checkpoint.
    """
    if meta is None:
        meta = {}
    elif not isinstance(meta, dict):
        raise TypeError(f'meta must be a dict or None, but got {type(meta)}')
    meta.update(mmcv_version=mmcv.__version__, time=time.asctime())

    if is_module_wrapper(model):
        model = model.module

    if hasattr(model, 'CLASSES') and model.CLASSES is not None:
        # save class name to the meta
        meta.update(CLASSES=model.CLASSES)

    checkpoint = {
        'meta': meta,
        'state_dict': weights_to_cpu(get_state_dict(model))
    }
    # save optimizer state dict in the checkpoint
    if isinstance(optimizer, Optimizer):
        checkpoint['optimizer'] = optimizer.state_dict()
    elif isinstance(optimizer, dict):
        checkpoint['optimizer'] = {}
        for name, optim in optimizer.items():
            checkpoint['optimizer'][name] = optim.state_dict()

    if filename.startswith('pavi://'):
        try:
            from pavi import modelcloud
            from pavi.exception import NodeNotFoundError
        except ImportError:
            raise ImportError(
                'Please install pavi to load checkpoint from modelcloud.')
        model_path = filename[7:]
        root = modelcloud.Folder()
        model_dir, model_name = osp.split(model_path)
        try:
            model = modelcloud.get(model_dir)
        except NodeNotFoundError:
            model = root.create_training_model(model_dir)
        with TemporaryDirectory() as tmp_dir:
            checkpoint_file = osp.join(tmp_dir, model_name)
            with open(checkpoint_file, 'wb') as f:
                torch.save(checkpoint, f)
                f.flush()
            model.create_file(checkpoint_file, name=model_name)
    else:
        mmcv.mkdir_or_exist(osp.dirname(filename))
        # immediately flush buffer
        with open(filename, 'wb') as f:
            torch.save(checkpoint, f)
            f.flush()



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.cuda.amp as amp



##
# version 1: use torch.autograd
class LabelSmoothSoftmaxCEV1(nn.Module):
    '''
    This is the autograd version, you can also try the LabelSmoothSoftmaxCEV2 that uses derived gradients
    '''

    def __init__(self, lb_smooth=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothSoftmaxCEV1, self).__init__()
        self.lb_smooth = lb_smooth
        self.reduction = reduction
        self.lb_ignore = ignore_index
        self.log_softmax = nn.LogSoftmax(dim=1)

    def forward(self, logits, label):
        '''
        Same usage method as nn.CrossEntropyLoss:
            >>> criteria = LabelSmoothSoftmaxCEV1()
            >>> logits = torch.randn(8, 19, 384, 384) # nchw, float/half
            >>> lbs = torch.randint(0, 19, (8, 384, 384)) # nhw, int64_t
            >>> loss = criteria(logits, lbs)
        '''
        # overcome ignored label
        logits = logits.float() # use fp32 to avoid nan
        with torch.no_grad():
            num_classes = logits.size(1)
            label = label.clone().detach()
            ignore = label.eq(self.lb_ignore)
            n_valid = ignore.eq(0).sum()
            label[ignore] = 0 # bhw
            lb_pos, lb_neg = 1. - self.lb_smooth, self.lb_smooth / num_classes
            lb_one_hot = torch.empty_like(logits).fill_(
                lb_neg).scatter_(1, label.unsqueeze(1), lb_pos).detach() # bchw

        logs = self.log_softmax(logits) # bchw
        loss = -torch.sum(logs * lb_one_hot, dim=1)
        loss[ignore] = 0
        if self.reduction == 'mean':
            loss = loss.sum() / n_valid
        if self.reduction == 'sum':
            loss = loss.sum()

        return loss



##
# version 2: user derived grad computation
class LSRCrossEntropyFunctionV2(torch.autograd.Function):

    @staticmethod
    @amp.custom_fwd(cast_inputs=torch.float32)
    def forward(ctx, logits, label, lb_smooth, lb_ignore):
        # prepare label
        num_classes = logits.size(1)
        lb_pos, lb_neg = 1. - lb_smooth, lb_smooth / num_classes
        label = label.clone().detach()
        ignore = label.eq(lb_ignore)
        n_valid = ignore.eq(0).sum()
        label[ignore] = 0
        lb_one_hot = torch.empty_like(logits).fill_(
            lb_neg).scatter_(1, label.unsqueeze(1), lb_pos).detach()

        ignore = ignore.nonzero(as_tuple=False)
        _, M = ignore.size()
        a, *b = ignore.chunk(M, dim=1)
        mask = [a, torch.arange(logits.size(1)), *b]
        lb_one_hot[mask] = 0
        coeff = (num_classes - 1) * lb_neg + lb_pos

        ctx.variables = coeff, mask, logits, lb_one_hot

        loss = torch.log_softmax(logits, dim=1).neg_().mul_(lb_one_hot).sum(dim=1)
        return loss

    @staticmethod
    @amp.custom_bwd
    def backward(ctx, grad_output):
        coeff, mask, logits, lb_one_hot = ctx.variables

        scores = torch.softmax(logits, dim=1).mul_(coeff)
        grad = scores.sub_(lb_one_hot).mul_(grad_output.unsqueeze(1))
        grad[mask] = 0
        return grad, None, None, None


class LabelSmoothSoftmaxCEV2(nn.Module):

    def __init__(self, lb_smooth=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothSoftmaxCEV2, self).__init__()
        self.lb_smooth = lb_smooth
        self.reduction = reduction
        self.lb_ignore = ignore_index

    def forward(self, logits, labels):
        '''
        Same usage method as nn.CrossEntropyLoss:
            >>> criteria = LabelSmoothSoftmaxCEV2()
            >>> logits = torch.randn(8, 19, 384, 384) # nchw, float/half
            >>> lbs = torch.randint(0, 19, (8, 384, 384)) # nhw, int64_t
            >>> loss = criteria(logits, lbs)
        '''
        losses = LSRCrossEntropyFunctionV2.apply(
                logits, labels, self.lb_smooth, self.lb_ignore)
        if self.reduction == 'sum':
            losses = losses.sum()
        elif self.reduction == 'mean':
            n_valid = (labels != self.lb_ignore).sum()
            losses = losses.sum() / n_valid
        return losses

# ##
# # version 3: implement wit cpp/cuda to save memory and accelerate
# import lsr_cpp
# class LSRCrossEntropyFunctionV3(torch.autograd.Function):
#     '''
#     use cpp/cuda to accelerate and shrink memory usage
#     '''
#     @staticmethod
#     @amp.custom_fwd(cast_inputs=torch.float32)
#     def forward(ctx, logits, labels, lb_smooth, lb_ignore):
#         losses = lsr_cpp.lsr_forward(logits, labels, lb_ignore, lb_smooth)

#         ctx.variables = logits, labels, lb_ignore, lb_smooth
#         return losses

#     @staticmethod
#     @amp.custom_bwd
#     def backward(ctx, grad_output):
#         logits, labels, lb_ignore, lb_smooth = ctx.variables

#         grad = lsr_cpp.lsr_backward(logits, labels, lb_ignore, lb_smooth)
#         grad.mul_(grad_output.unsqueeze(1))
#         return grad, None, None, None


class LabelSmoothSoftmaxCEV3(nn.Module):

    def __init__(self, lb_smooth=0.1, reduction='mean', ignore_index=-100):
        super(LabelSmoothSoftmaxCEV3, self).__init__()
        self.lb_smooth = lb_smooth
        self.reduction = reduction
        self.lb_ignore = ignore_index

    def forward(self, logits, labels):
        '''
        Same usage method as nn.CrossEntropyLoss:
            >>> criteria = LabelSmoothSoftmaxCEV3()
            >>> logits = torch.randn(8, 19, 384, 384) # nchw, float/half
            >>> lbs = torch.randint(0, 19, (8, 384, 384)) # nhw, int64_t
            >>> loss = criteria(logits, lbs)
        '''
        losses = LSRCrossEntropyFunctionV3.apply(
                logits, labels, self.lb_smooth, self.lb_ignore)
        if self.reduction == 'sum':
            losses = losses.sum()
        elif self.reduction == 'mean':
            n_valid = (labels != self.lb_ignore).sum()
            losses = losses.sum() / n_valid
        return losses


if __name__ == '__main__':
    import torchvision
    import torch
    import numpy as np
    import random
    torch.manual_seed(15)
    random.seed(15)
    np.random.seed(15)
    torch.backends.cudnn.deterministic = True

    class Model(nn.Module):
        def __init__(self):
            super(Model, self).__init__()
            net = torchvision.models.resnet18(pretrained=False)
            self.conv1 = net.conv1
            self.bn1 = net.bn1
            self.maxpool = net.maxpool
            self.relu = net.relu
            self.layer1 = net.layer1
            self.layer2 = net.layer2
            self.layer3 = net.layer3
            self.layer4 = net.layer4
            self.fc = nn.Conv2d(512, 19, 3, 1, 1)
        def forward(self, x):
            feat = self.conv1(x)
            feat = self.bn1(feat)
            feat = self.relu(feat)
            feat = self.maxpool(feat)
            feat = self.layer1(feat)
            feat = self.layer2(feat)
            feat = self.layer3(feat)
            feat = self.layer4(feat)
            feat = self.fc(feat)
            out = F.interpolate(feat, x.size()[2:], mode='bilinear', align_corners=True)
            return out

    net1 = Model()
    net2 = Model()
    net2.load_state_dict(net1.state_dict())
    red = 'mean'
    criteria1 = LabelSmoothSoftmaxCEV2(lb_smooth=0.1, ignore_index=255, reduction=red)
    criteria2 = LabelSmoothSoftmaxCEV1(lb_smooth=0.1, ignore_index=255, reduction=red)
    net1.cuda()
    net2.cuda()
    net1.train()
    net2.train()
    criteria1.cuda()
    criteria2.cuda()

    optim1 = torch.optim.SGD(net1.parameters(), lr=1e-2)
    optim2 = torch.optim.SGD(net2.parameters(), lr=1e-2)

    bs = 64
    for it in range(300):
        inten = torch.randn(bs, 3, 224, 224).cuda()
        lbs = torch.randint(0, 19, (bs, 224, 224)).cuda()
        lbs[1, 1, 1] = 255
        lbs[30, 3, 2:200] = 255
        lbs[18, 4:7, 8:200] = 255
        logits = net1(inten)
        loss1 = criteria1(logits, lbs)
        optim1.zero_grad()
        loss1.backward()
        optim1.step()
        #  print(net1.fc.weight[:, :5])
        logits = net2(inten)
        loss2 = criteria2(logits, lbs)
        optim2.zero_grad()
        loss2.backward()
        optim2.step()
        #  net1.load_state_dict(net2.state_dict())
        #  print(net2.fc.weight[:, :5])
        with torch.no_grad():
            if (it+1) % 50 == 0:
                print('iter: {}, ================='.format(it+1))
                #  print(net1.fc.weight.numel())
                print('fc weight: ', torch.mean(torch.abs(net1.fc.weight - net2.fc.weight)).item())

                print('conv1 weight: ', torch.mean(torch.abs(net1.conv1.weight - net2.conv1.weight)).item())
                print('loss: ', loss1.item() - loss2.item())


import glob 
import os
import os.path as osp 
import json 
from tqdm import tqdm


def labelme2yolo_instance_segmentation(input_dir, output_dir, image_ext,
                                       copy_image=True, image_width=None, image_height=None):

    if not osp.exists(output_dir):
        os.mkdir(output_dir)
        
    folders = [folder.split("/")[-1] for folder in glob.glob(osp.join(input_dir, "**")) if not osp.isfile(folder)]


    class2idx = {}
    for folder in folders:
        _output_labels_dir = osp.join(output_dir, 'labels', folder)
        if not osp.exists(_output_labels_dir):
            os.makedirs(_output_labels_dir)

        json_files = glob.glob(osp.join(input_dir, folder, '*.json'))
        print(f"There are {len(json_files)} json files")

        for json_file in tqdm(json_files, desc=folder):
            filename = osp.split(osp.splitext(json_file)[0])[-1]
            txt = open(osp.join(_output_labels_dir, filename + '.txt'), 'w')
            with open(json_file, 'r') as jf:
                anns = json.load(jf)['shapes']
                
            if copy_image:
                import cv2
                from shutil import copyfile
                img_file = osp.join(input_dir, folder, filename + f'.{image_ext}')
                assert osp.exists(img_file), ValueError(f"There is no such image: {img_file}")
                
                image = cv2.imread(img_file)
                image_width = image.shape[1]
                image_height = image.shape[0]
                _output_image_dir = osp.join(output_dir, 'images', folder)
                if not osp.exists(_output_image_dir):
                    os.makedirs(_output_image_dir)
                copyfile(img_file, osp.join(_output_image_dir, filename + f'.{image_ext}'))
                
                
            assert image_width != None and image_height != None, ValueError(f"The input size (width, height) must be assigned")
                
            if len(anns) != 0:
                for ann in anns:
                    shape_type = ann['shape_type']
                    label = ann['label']
                    if label not in class2idx.keys():
                        class2idx.update({label: len(class2idx)})
                    points = ann['points']
                    if shape_type == 'point' or len(points) <= 2:
                        continue
                    assert len(points) >= 3, RuntimeError(f"The number of polygon points must be more than 3, not {len(points)} with {shape_type}")

                    txt.write(str(class2idx[label]) + ' ')
                    for idx, point in enumerate(points):
                        if idx == len(points) -1:
                            txt.write(f'{round(point[0]/image_width, 3)} {round(point[1]/image_height, 3)}')
                        else:
                            txt.write(f'{round(point[0]/image_width, 3)} {round(point[1]/image_height, 3)} ')
                    txt.write('\n')
            
        txt.close()

    txt = open(osp.join(output_dir, 'classes.txt'), 'w')
    for key, val in class2idx.items():
        txt.write(f'{val}: {key}\n')
    txt.close()
            
            
            
if __name__ == '__main__':
    input_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/inner_body/split_dataset'
    output_dir = '/DeepLearning/etc/_athena_tests/benchmark/sungwoo/inner_body/split_dataset_yolo_is'

    copy_image = True
    image_ext = 'bmp'
    
    image_width = None
    image_height = None
    
    labelme2yolo_instance_segmentation(input_dir, output_dir, image_ext, copy_image, image_width, image_height)


from abc import abstractmethod
from visionsuite.engines.classification.utils.registry import DATASETS
from visionsuite.engines.utils.bases.base_oop_module import BaseOOPModule


@DATASETS.register()
class BaseDataset(BaseOOPModule):
    def __init__(self, name=None, transform=None, mode='train'):
        super().__init__(name=name)
        self.args = None
        
        self._transform = transform
        self.train_dataset = None
        self.val_dataset = None
        
        self.train_sampler = None 
        self.val_sampler = None
        
        self.label2index = None
        self.index2label = None
        self.num_classes = None
        self.classes = None
        
    @property 
    def transform(self):
        return self._transform 
    
    def build(self, load=True, *args, **kwargs):
        super().build(*args, **kwargs)
        
        if 'classes' in kwargs:
            self.classes = kwargs['classes']
            
        self.log_info(f"Built dataset: {self.args}", self.build.__name__, __class__.__name__)
        
        if load:
            self._load()
    
    @BaseOOPModule.track_status
    def _load(self):
        self.load_dataset()
        self.load_sampler()

    @BaseOOPModule.track_status
    @abstractmethod
    def load_dataset(self):
        self.log_info(f"Loading dataset", self.load_dataset.__name__, __class__.__name__)
        pass

    @BaseOOPModule.track_status
    def load_sampler(self):
        self.log_info(f"Loading sampler", self.load_sampler.__name__, __class__.__name__)
        import torch 
        from visionsuite.engines.classification.utils.registry import SAMPLERS

        if self.args['distributed']['use']:
            if self.args['sampler']['type']:
                self.train_sampler = SAMPLERS.get(self.args['sampler']['type'])(self.train_dataset, shuffle=True, repetitions=self.args['sampler']['reps'])
            else:
                self.train_sampler = torch.utils.data.distributed.DistributedSampler(self.train_dataset)
            self.val_sampler = torch.utils.data.distributed.DistributedSampler(self.val_dataset, shuffle=False)
        else:
            self.train_sampler = torch.utils.data.RandomSampler(self.train_dataset)
            self.val_sampler = torch.utils.data.SequentialSampler(self.val_dataset)
            

# def load_data(traindir, valdir, transform, args):
#     import os
#     import torch
#     import torchvision
#     import time
#     import torchvision.transforms
#     from visionsuite.engines.utils.helpers import mkdir, get_cache_path
#     from visionsuite.engines.utils.torch_utils.utils import save_on_master
#     from visionsuite.engines.classification.utils.registry import SAMPLERS

#     # Data loading code
#     print("Loading data")
    

#     st = time.time()
#     cache_path = get_cache_path(traindir)
#     if args.cache_dataset and os.path.exists(cache_path):
#         # Attention, as the transforms are also cached!
#         print(f"Loading dataset_train from {cache_path}")
#         # TODO: this could probably be weights_only=True
#         dataset, _ = torch.load(cache_path, weights_only=False)
#     else:
#         dataset = torchvision.datasets.ImageFolder(
#             traindir,
#             transform,
#         )
#         if args.cache_dataset:
#             print(f"Saving dataset_train to {cache_path}")
#             mkdir(os.path.dirname(cache_path))
#             save_on_master((dataset, traindir), cache_path)
#     print("Took", time.time() - st)

#     print("Loading validation data")
#     cache_path = get_cache_path(valdir)
#     if args.cache_dataset and os.path.exists(cache_path):
#         # Attention, as the transforms are also cached!
#         print(f"Loading dataset_test from {cache_path}")
#         # TODO: this could probably be weights_only=True
#         dataset_test, _ = torch.load(cache_path, weights_only=False)
#     else:
#         if args.model['weights']:
#             weights = torchvision.models.get_weight(args.model['weights'])
#             preprocessing = weights.transforms(antialias=True)
#             if args.backend == "tensor":
#                 preprocessing = torchvision.transforms.Compose([torchvision.transforms.PILToTensor(), preprocessing])

#         else:
#             preprocessing = transform

#         dataset_test = torchvision.datasets.ImageFolder(
#             valdir,
#             preprocessing,
#         )
#         if args.cache_dataset:
#             print(f"Saving dataset_test to {cache_path}")
#             mkdir(os.path.dirname(cache_path))
#             save_on_master((dataset_test, valdir), cache_path)

#     print("Creating data loaders")


#     train_sampler, test_sampler = SAMPLERS.get('get_samplers')(args, dataset, dataset_test)

#     return dataset, dataset_test, train_sampler, test_sampler





from __future__ import absolute_import

from tensorflow.keras.applications import *
from tensorflow.keras.models import Model

from keras_unet_collection.utils import freeze_model

import warnings

layer_cadidates = {
    'VGG16': ('block1_conv2', 'block2_conv2', 'block3_conv3', 'block4_conv3', 'block5_conv3'),
    'VGG19': ('block1_conv2', 'block2_conv2', 'block3_conv4', 'block4_conv4', 'block5_conv4'),
    'ResNet50': ('conv1_relu', 'conv2_block3_out', 'conv3_block4_out', 'conv4_block6_out', 'conv5_block3_out'),
    'ResNet101': ('conv1_relu', 'conv2_block3_out', 'conv3_block4_out', 'conv4_block23_out', 'conv5_block3_out'),
    'ResNet152': ('conv1_relu', 'conv2_block3_out', 'conv3_block8_out', 'conv4_block36_out', 'conv5_block3_out'),
    'ResNet50V2': ('conv1_conv', 'conv2_block3_1_relu', 'conv3_block4_1_relu', 'conv4_block6_1_relu', 'post_relu'),
    'ResNet101V2': ('conv1_conv', 'conv2_block3_1_relu', 'conv3_block4_1_relu', 'conv4_block23_1_relu', 'post_relu'),
    'ResNet152V2': ('conv1_conv', 'conv2_block3_1_relu', 'conv3_block8_1_relu', 'conv4_block36_1_relu', 'post_relu'),
    'DenseNet121': ('conv1/relu', 'pool2_conv', 'pool3_conv', 'pool4_conv', 'relu'),
    'DenseNet169': ('conv1/relu', 'pool2_conv', 'pool3_conv', 'pool4_conv', 'relu'),
    'DenseNet201': ('conv1/relu', 'pool2_conv', 'pool3_conv', 'pool4_conv', 'relu'),
    'EfficientNetB0': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB1': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB2': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB3': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB4': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB5': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB6': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),
    'EfficientNetB7': ('block2a_expand_activation', 'block3a_expand_activation', 'block4a_expand_activation', 'block6a_expand_activation', 'top_activation'),}

def bach_norm_checker(backbone_name, batch_norm):
    '''batch norm checker'''
    if 'VGG' in backbone_name:
        batch_norm_backbone = False
    else:
        batch_norm_backbone = True
        
    if batch_norm_backbone != batch_norm:       
        if batch_norm_backbone:    
            param_mismatch = "\n\nBackbone {} uses batch norm, but other layers received batch_norm={}".format(backbone_name, batch_norm)
        else:
            param_mismatch = "\n\nBackbone {} does not use batch norm, but other layers received batch_norm={}".format(backbone_name, batch_norm)
            
        warnings.warn(param_mismatch);
        
def backbone_zoo(backbone_name, weights, input_tensor, depth, freeze_backbone, freeze_batch_norm):
    '''
    Configuring a user specified encoder model based on the `tensorflow.keras.applications`
    
    Input
    ----------
        backbone_name: the bakcbone model name. Expected as one of the `tensorflow.keras.applications` class.
                       Currently supported backbones are:
                       (1) VGG16, VGG19
                       (2) ResNet50, ResNet101, ResNet152
                       (3) ResNet50V2, ResNet101V2, ResNet152V2
                       (4) DenseNet121, DenseNet169, DenseNet201
                       (5) EfficientNetB[0,7]
                       
        weights: one of None (random initialization), 'imagenet' (pre-training on ImageNet), 
                 or the path to the weights file to be loaded.
        input_tensor: the input tensor 
        depth: number of encoded feature maps. 
               If four dwonsampling levels are needed, then depth=4.
        
        freeze_backbone: True for a frozen backbone
        freeze_batch_norm: False for not freezing batch normalization layers.
        
    Output
    ----------
        model: a keras backbone model.
        
    '''
    
    cadidate = layer_cadidates[backbone_name]
    
    # ----- #
    # depth checking
    depth_max = len(cadidate)
    if depth > depth_max:
        depth = depth_max
    # ----- #
    
    backbone_func = eval(backbone_name)
    backbone_ = backbone_func(include_top=False, weights=weights, input_tensor=input_tensor, pooling=None,)
    
    X_skip = []
    
    for i in range(depth):
        X_skip.append(backbone_.get_layer(cadidate[i]).output)
        
    model = Model(inputs=[input_tensor,], outputs=X_skip, name='{}_backbone'.format(backbone_name))
    
    if freeze_backbone:
        model = freeze_model(model, freeze_batch_norm=freeze_batch_norm)
    
    return model




from __future__ import absolute_import

from keras_unet_collection.layer_utils import *
from keras_unet_collection.activations import GELU, Snake
from keras_unet_collection._backbone_zoo import backbone_zoo, bach_norm_checker
from keras_unet_collection._model_unet_2d import UNET_left, UNET_right

from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

import warnings

def unet_plus_2d_base(input_tensor, filter_num, stack_num_down=2, stack_num_up=2,
                      activation='ReLU', batch_norm=False, pool=True, unpool=True, deep_supervision=False, 
                      backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='xnet'):
    '''
    The base of U-net++ with an optional ImageNet-trained backbone
    
    unet_plus_2d_base(input_tensor, filter_num, stack_num_down=2, stack_num_up=2,
                      activation='ReLU', batch_norm=False, pool=True, unpool=True, deep_supervision=False, 
                      backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='xnet')
    
    ----------
    Zhou, Z., Siddiquee, M.M.R., Tajbakhsh, N. and Liang, J., 2018. Unet++: A nested u-net architecture 
    for medical image segmentation. In Deep Learning in Medical Image Analysis and Multimodal Learning 
    for Clinical Decision Support (pp. 3-11). Springer, Cham.
    
    Input
    ----------
        input_tensor: the input tensor of the base, e.g., `keras.layers.Inpyt((None, None, 3))`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.   
        deep_supervision: True for a model that supports deep supervision. Details see Zhou et al. (2018).
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of backbone options) ----------
        backbone_name: the bakcbone model name. Should be one of the `tensorflow.keras.applications` class.
                       None (default) means no backbone. 
                       Currently supported backbones are:
                       (1) VGG16, VGG19
                       (2) ResNet50, ResNet101, ResNet152
                       (3) ResNet50V2, ResNet101V2, ResNet152V2
                       (4) DenseNet121, DenseNet169, DenseNet201
                       (5) EfficientNetB[0-7]
        weights: one of None (random initialization), 'imagenet' (pre-training on ImageNet), 
                 or the path to the weights file to be loaded.
        freeze_backbone: True for a frozen backbone.
        freeze_batch_norm: False for not freezing batch normalization layers.
        
    Output
    ----------
        If deep_supervision = False; Then the output is a tensor.
        If deep_supervision = True; Then the output is a list of tensors
            with the first tensor obtained from the first downsampling level (for checking the input/output shapes only),
            the second to the `depth-1`-th tensors obtained from each intermediate upsampling levels (deep supervision tensors),
            and the last tensor obtained from the end of the base.
    
    '''
    
    activation_func = eval(activation)

    depth_ = len(filter_num)
    # allocate nested lists for collecting output tensors 
    X_nest_skip = [[] for _ in range(depth_)]

    # no backbone cases
    if backbone is None:

        X = input_tensor

        # downsampling blocks (same as in 'unet_2d')
        X = CONV_stack(X, filter_num[0], stack_num=stack_num_down, activation=activation, 
                       batch_norm=batch_norm, name='{}_down0'.format(name))
        X_nest_skip[0].append(X)
        for i, f in enumerate(filter_num[1:]):
            X = UNET_left(X, f, stack_num=stack_num_down, activation=activation, 
                          pool=pool, batch_norm=batch_norm, name='{}_down{}'.format(name, i+1))        
            X_nest_skip[0].append(X)

    # backbone cases
    else:        
        # handling VGG16 and VGG19 separately
        if 'VGG' in backbone:
            backbone_ = backbone_zoo(backbone, weights, input_tensor, depth_, freeze_backbone, freeze_batch_norm)
            # collecting backbone feature maps
            X_nest_skip[0] += backbone_([input_tensor,])
            depth_encode = len(X_nest_skip[0])

        # for other backbones
        else:
            backbone_ = backbone_zoo(backbone, weights, input_tensor, depth_-1, freeze_backbone, freeze_batch_norm)
            # collecting backbone feature maps
            X_nest_skip[0] += backbone_([input_tensor,])
            depth_encode = len(X_nest_skip[0]) + 1

        # extra conv2d blocks are applied
        # if downsampling levels of a backbone < user-specified downsampling levels
        if depth_encode < depth_:

            # begins at the deepest available tensor  
            X = X_nest_skip[0][-1]

            # extra downsamplings
            for i in range(depth_-depth_encode):
                i_real = i + depth_encode

                X = UNET_left(X, filter_num[i_real], stack_num=stack_num_down, activation=activation, pool=pool, 
                              batch_norm=batch_norm, name='{}_down{}'.format(name, i_real+1))
                X_nest_skip[0].append(X)


    X = X_nest_skip[0][-1]

    for nest_lev in range(1, depth_):

        # depth difference between the deepest nest skip and the current upsampling  
        depth_lev = depth_-nest_lev

        # number of available encoded tensors
        depth_decode = len(X_nest_skip[nest_lev-1])

        # loop over individual upsamling levels
        for i in range(1, depth_decode):

            # collecting previous downsampling outputs
            previous_skip = []
            for previous_lev in range(nest_lev):
                previous_skip.append(X_nest_skip[previous_lev][i-1])

            # upsamping block that concatenates all available (same feature map size) down-/upsampling outputs
            X_nest_skip[nest_lev].append(
                UNET_right(X_nest_skip[nest_lev-1][i], previous_skip, filter_num[i-1], 
                           stack_num=stack_num_up, activation=activation, unpool=unpool, 
                           batch_norm=batch_norm, concat=False, name='{}_up{}_from{}'.format(name, nest_lev-1, i-1)))

        if depth_decode < depth_lev+1:

            X = X_nest_skip[nest_lev-1][-1]

            for j in range(depth_lev-depth_decode+1):
                j_real = j + depth_decode
                X = UNET_right(X, None, filter_num[j_real-1], 
                               stack_num=stack_num_up, activation=activation, unpool=unpool, 
                               batch_norm=batch_norm, concat=False, name='{}_up{}_from{}'.format(name, nest_lev-1, j_real-1))
                X_nest_skip[nest_lev].append(X)
            
    # output
    if deep_supervision:
        
        X_list = []
        
        for i in range(depth_):
            X_list.append(X_nest_skip[i][0])
        
        return X_list
        
    else:
        return X_nest_skip[-1][0]

def unet_plus_2d(input_size, filter_num, n_labels, stack_num_down=2, stack_num_up=2,
                 activation='ReLU', output_activation='Softmax', batch_norm=False, pool=True, unpool=True, deep_supervision=False, 
                 backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='xnet'):
    '''
    U-net++ with an optional ImageNet-trained backbone.
    
    unet_plus_2d(input_size, filter_num, n_labels, stack_num_down=2, stack_num_up=2,
                 activation='ReLU', output_activation='Softmax', batch_norm=False, pool=True, unpool=True, deep_supervision=False, 
                 backbone=None, weights='imagenet', freeze_backbone=True, freeze_batch_norm=True, name='xnet')
    
    ----------
    Zhou, Z., Siddiquee, M.M.R., Tajbakhsh, N. and Liang, J., 2018. Unet++: A nested u-net architecture 
    for medical image segmentation. In Deep Learning in Medical Image Analysis and Multimodal Learning 
    for Clinical Decision Support (pp. 3-11). Springer, Cham.
    
    Input
    ----------
        input_size: the size/shape of network input, e.g., `(128, 128, 3)`.
        filter_num: a list that defines the number of filters for each \
                    down- and upsampling levels. e.g., `[64, 128, 256, 512]`.
                    The depth is expected as `len(filter_num)`.
        n_labels: number of output labels.
        stack_num_down: number of convolutional layers per downsampling level/block. 
        stack_num_up: number of convolutional layers (after concatenation) per upsampling level/block.
        activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interfaces, e.g., 'ReLU'.
        output_activation: one of the `tensorflow.keras.layers` or `keras_unet_collection.activations` interface or 'Sigmoid'.
                           Default option is 'Softmax'.
                           if None is received, then linear activation is applied.
        batch_norm: True for batch normalization.
        pool: True or 'max' for MaxPooling2D.
              'ave' for AveragePooling2D.
              False for strided conv + batch norm + activation.
        unpool: True or 'bilinear' for Upsampling2D with bilinear interpolation.
                'nearest' for Upsampling2D with nearest interpolation.
                False for Conv2DTranspose + batch norm + activation.   
        deep_supervision: True for a model that supports deep supervision. Details see Zhou et al. (2018).
        name: prefix of the created keras model and its layers.
        
        ---------- (keywords of backbone options) ----------
        backbone_name: the bakcbone model name. Should be one of the `tensorflow.keras.applications` class.
                       None (default) means no backbone. 
                       Currently supported backbones are:
                       (1) VGG16, VGG19
                       (2) ResNet50, ResNet101, ResNet152
                       (3) ResNet50V2, ResNet101V2, ResNet152V2
                       (4) DenseNet121, DenseNet169, DenseNet201
                       (5) EfficientNetB[0-7]
        weights: one of None (random initialization), 'imagenet' (pre-training on ImageNet), 
                 or the path to the weights file to be loaded.
        freeze_backbone: True for a frozen backbone.
        freeze_batch_norm: False for not freezing batch normalization layers.
        
    Output
    ----------
        model: a keras model.
    
    '''
    
    depth_ = len(filter_num)
    
    if backbone is not None:
        bach_norm_checker(backbone, batch_norm)
    
    IN = Input(input_size)
    # base
    X = unet_plus_2d_base(IN, filter_num, stack_num_down=stack_num_down, stack_num_up=stack_num_up,
                          activation=activation, batch_norm=batch_norm, pool=pool, unpool=unpool, deep_supervision=deep_supervision, 
                          backbone=backbone, weights=weights, freeze_backbone=freeze_backbone, freeze_batch_norm=freeze_batch_norm, name=name)
    
    # output
    if deep_supervision:
        
        if (backbone is not None) and freeze_backbone:
            backbone_warn = '\n\nThe shallowest U-net++ deep supervision branch directly connects to a frozen backbone.\nTesting your configurations on `keras_unet_collection.base.unet_plus_2d_base` is recommended.'
            warnings.warn(backbone_warn);
            
        # model base returns a list of tensors
        X_list = X
        OUT_list = []
        
        print('----------\ndeep_supervision = True\nnames of output tensors are listed as follows ("sup0" is the shallowest supervision layer;\n"final" is the final output layer):\n')
        
        # no backbone or VGG backbones
        # depth_ > 2 is expected (a least two downsampling blocks)
        if (backbone is None) or 'VGG' in backbone:
        
            for i in range(0, depth_-1):
                if output_activation is None:
                    print('\t{}_output_sup{}'.format(name, i))
                else:
                    print('\t{}_output_sup{}_activation'.format(name, i))
                    
                OUT_list.append(CONV_output(X_list[i], n_labels, kernel_size=1, activation=output_activation, 
                                            name='{}_output_sup{}'.format(name, i)))
        # other backbones        
        else:
            for i in range(1, depth_-1):
                if output_activation is None:
                    print('\t{}_output_sup{}'.format(name, i-1))
                else:
                    print('\t{}_output_sup{}_activation'.format(name, i-1))
                
                # an extra upsampling for creating full resolution feature maps
                X = decode_layer(X_list[i], filter_num[i], 2, unpool, activation=activation, 
                                 batch_norm=batch_norm, name='{}_sup{}_up'.format(name, i-1))
                
                X = CONV_output(X, n_labels, kernel_size=1, activation=output_activation, name='{}_output_sup{}'.format(name, i-1))
                OUT_list.append(X)
                
        if output_activation is None:
            print('\t{}_output_final'.format(name))
        else:
            print('\t{}_output_final_activation'.format(name))
            
        OUT_list.append(CONV_output(X_list[-1], n_labels, kernel_size=1, activation=output_activation, name='{}_output_final'.format(name)))
        
    else:
        OUT = CONV_output(X, n_labels, kernel_size=1, activation=output_activation, name='{}_output'.format(name))
        OUT_list = [OUT,]
        
    # model
    model = Model(inputs=[IN,], outputs=OUT_list, name='{}_model'.format(name))
    
    return model



# Copyright (c) OpenMMLab. All rights reserved.
from .bbox_heads import (RotatedBBoxHead, RotatedConvFCBBoxHead,
                         RotatedShared2FCBBoxHead)
from mmrotate.models.roi_heads.gv_ratio_roi_head import GVRatioRoIHead
from mmrotate.models.roi_heads.oriented_standard_roi_head import OrientedStandardRoIHead
from mmrotate.models.roi_heads.roi_extractors import RotatedSingleRoIExtractor
from mmrotate.models.roi_heads.roi_trans_roi_head import RoITransRoIHead
from .custom_rotate_standard_roi_head import CustomRotatedStandardRoIHead

from .rotate_standard_roi_head_imted import RotatedStandardRoIHeadimTED
from .oriented_standard_roi_head_imted import OrientedStandardRoIHeadimTED

__all__ = [
    'RotatedBBoxHead', 'RotatedConvFCBBoxHead', 'RotatedShared2FCBBoxHead',
    'CustomRotatedStandardRoIHead', 'RotatedSingleRoIExtractor',
    'OrientedStandardRoIHead', 'RoITransRoIHead', 'GVRatioRoIHead',

    'RotatedStandardRoIHeadimTED',
    'OrientedStandardRoIHeadimTED',
]



import numpy as np

MEAN = (0.485, 0.456, 0.406)
STD = (0.229, 0.224, 0.225)

def denormalize(x, mean=MEAN, std=STD):
    x *= np.array(std)
    x += np.array(mean)
    x *= 255
    x = x.astype(np.uint8)
    
    return x


import copy
import os

import torch
import torch.utils.data
import torchvision
from PIL import Image
from pycocotools import mask as coco_mask
from visionsuite.engines.utils.torch_utils.transforms import Compose


from visionsuite.engines.segmentation.utils.registry import DATASETS
from visionsuite.engines.segmentation.src.datasets.base_dataset import BaseDataset

import torchvision
from torchvision.transforms import functional as F, InterpolationMode
import visionsuite.engines.utils.torch_utils.presets as presets

def get_transform(is_train, args):
    if is_train:
        return presets.SegmentationPresetTrain(base_size=512, crop_size=480, backend=args['backend'], use_v2=args['use_v2'])
    elif args['weights'] and args['test_only']:
        weights = torchvision.models.get_weight(args['weights'])
        trans = weights.transforms()

        def preprocessing(img, target):
            img = trans(img)
            size = F.get_dimensions(img)[1:]
            target = F.resize(target, size, interpolation=InterpolationMode.NEAREST)
            return img, F.pil_to_tensor(target)

        return preprocessing
    else:
        return presets.SegmentationPresetEval(base_size=512, backend=args['backend'], use_v2=args['use_v2'])

import copy
import os

import torch
import torch.utils.data
import torchvision
from PIL import Image
from pycocotools import mask as coco_mask
from visionsuite.engines.utils.torch_utils.transforms import Compose


class FilterAndRemapCocoCategories:
    def __init__(self, categories, remap=True):
        self.categories = categories
        self.remap = remap

    def __call__(self, image, anno):
        anno = [obj for obj in anno if obj["category_id"] in self.categories]
        if not self.remap:
            return image, anno
        anno = copy.deepcopy(anno)
        for obj in anno:
            obj["category_id"] = self.categories.index(obj["category_id"])
        return image, anno


def convert_coco_poly_to_mask(segmentations, height, width):
    masks = []
    for polygons in segmentations:
        rles = coco_mask.frPyObjects(polygons, height, width)
        mask = coco_mask.decode(rles)
        if len(mask.shape) < 3:
            mask = mask[..., None]
        mask = torch.as_tensor(mask, dtype=torch.uint8)
        mask = mask.any(dim=2)
        masks.append(mask)
    if masks:
        masks = torch.stack(masks, dim=0)
    else:
        masks = torch.zeros((0, height, width), dtype=torch.uint8)
    return masks


class ConvertCocoPolysToMask:
    def __call__(self, image, anno):
        w, h = image.size
        segmentations = [obj["segmentation"] for obj in anno]
        cats = [obj["category_id"] for obj in anno]
        if segmentations:
            masks = convert_coco_poly_to_mask(segmentations, h, w)
            cats = torch.as_tensor(cats, dtype=masks.dtype)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            target, _ = (masks * cats[:, None, None]).max(dim=0)
            # discard overlapping instances
            target[masks.sum(0) > 1] = 255
        else:
            target = torch.zeros((h, w), dtype=torch.uint8)
        target = Image.fromarray(target.numpy())
        return image, target


def _coco_remove_images_without_annotations(dataset, cat_list=None):
    def _has_valid_annotation(anno):
        # if it's empty, there is no annotation
        if len(anno) == 0:
            return False
        # if more than 1k pixels occupied in the image
        return sum(obj["area"] for obj in anno) > 1000

    ids = []
    for ds_idx, img_id in enumerate(dataset.ids):
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        if cat_list:
            anno = [obj for obj in anno if obj["category_id"] in cat_list]
        if _has_valid_annotation(anno):
            ids.append(ds_idx)

    dataset = torch.utils.data.Subset(dataset, ids)
    return dataset



def get_dataset(root, image_set, transforms, use_v2=False):
    
    PATHS = {
                "train": ("train2017", os.path.join("annotations", "instances_train2017.json")),
                "val": ("val2017", os.path.join("annotations", "instances_val2017.json")),
                # "train": ("val2017", os.path.join("annotations", "instances_val2017.json"))
            }
    # CAT_LIST = [0, 5, 2, 16, 9, 44, 6, 3, 17, 62, 21, 67, 18, 19, 4, 1, 64, 20, 63, 7, 72]
    CAT_LIST = [0, 5, 2]

    img_folder, ann_file = PATHS[image_set]
    img_folder = os.path.join(root, img_folder)
    ann_file = os.path.join(root, ann_file)

    # The 2 "Compose" below achieve the same thing: converting coco detection
    # samples into segmentation-compatible samples. They just do it with
    # slightly different implementations. We could refactor and unify, but
    # keeping them separate helps keeping the v2 version clean
    if use_v2:
        import visionsuite.engines.utils.torch_utils.v2_extras as v2_extras
        from torchvision.datasets import wrap_dataset_for_transforms_v2

        transforms = Compose([v2_extras.CocoDetectionToVOCSegmentation(), transforms])
        dataset = torchvision.datasets.CocoDetection(img_folder, ann_file, transforms=transforms)
        dataset = wrap_dataset_for_transforms_v2(dataset, target_keys={"masks", "labels"})
    else:
        transforms = Compose([FilterAndRemapCocoCategories(CAT_LIST, remap=True), ConvertCocoPolysToMask(), transforms])
        dataset = torchvision.datasets.CocoDetection(img_folder, ann_file, transforms=transforms)

    if image_set == "train":
        dataset = _coco_remove_images_without_annotations(dataset, CAT_LIST)

    return dataset, CAT_LIST

@DATASETS.register()
class COCODataset(BaseDataset):
    def __init__(self, transform=None):
        super().__init__(transform=transform)
        
    def load_dataset(self):
        super().load_dataset()
        self.train_dataset, self.classes = get_dataset(self.args['input_dir'], 'train', 
                                         get_transform(True, {'weights': None, 'test_only': False, 'backend': 'PIL', 'use_v2': False}))
        self.val_dataset, _ = get_dataset(self.args['input_dir'], 'val', 
                                         get_transform(True, {'weights': None, 'test_only': False, 'backend': 'PIL', 'use_v2': False}))
        
        
        self.label2index = {index: label for index, label in enumerate(self.classes)}
        self.index2label = {label: index for index, label in enumerate(self.classes)}
        self.num_classes = len(self.classes)
        print(f"label2index: {self.label2index}")



import os.path as osp
import os

from src.test_obb import test_obb

if __name__ == '__main__':
    
    model_name = 'yolov8'
    backbone = 'l_dfl_0.2_norad'
    weights_file = f"/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/outputs/dfl_0.2_norad/weights/best.pt"

    # input_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/split_dataset_doosan/val'
    # json_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/split_dataset_doosan/val'
    # output_dir = f'/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/tests/doosan/{model_name}_{backbone}'
    input_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/sfaw'
    json_dir = '/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/dataset/sfaw'
    output_dir = f'/HDD/_projects/benchmark/obb_detection/doosan_cj_rich/tests/sfaw/{model_name}_{backbone}'
    
    if not osp.exists(output_dir):
        os.makedirs(output_dir)
    
    compare_gt = True
    iou_threshold = 0.9
    conf_threshold = 0.1
    line_width = 3
    font_scale = 2
    imgsz = 768
    _classes = ['BOX']
    input_img_ext = 'bmp'
    output_img_ext = 'jpg'
    output_img_size_ratio = 1

    test_obb(weights_file, imgsz, _classes, input_dir, output_dir, json_dir, compare_gt, 
                iou_threshold, conf_threshold, line_width, font_scale, 
                input_img_ext=input_img_ext, output_img_ext=output_img_ext, output_img_size_ratio=output_img_size_ratio)


import torch

def get_accuracies(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.inference_mode():
        maxk = max(topk)
        batch_size = target.size(0)
        if target.ndim == 2:
            target = target.max(dim=1)[1]

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target[None])

        res = []
        for k in topk:
            correct_k = correct[:k].flatten().sum(dtype=torch.float32)
            res.append(correct_k * (100.0 / batch_size))
        return res


import numpy as np

from visionsuite.engines.data.slicer.utils.functional import (
    finetune_patch_for_roi,
    get_center_point_from_points,
    get_points_in_patch,
    get_roi_image,
    is_point_in_roi,
)

from .sliding_patches_v2 import get_sliding_patches_v2


def get_centric_patches(
    points,
    patch_width,
    patch_height,
    img_width,
    img_height,
    roi=[],
    include_point_positive=False,
    filename=None,
    logger=None,
):

    # TODO: Need to consider number of pixels overlapped b/t patch and points b/c of the finetuning?

    [tl_x, tl_y, br_x, br_y] = get_roi_image(roi, img_width, img_height)

    if logger is not None:
        logger.assertion_log(
            patch_width <= br_x - tl_x,
            RuntimeError(f"patch width({patch_width}) should be bigger than width({br_x - tl_x})"),
            parent_fn=get_centric_patches.__name__,
        )
        logger.assertion_log(
            patch_height <= br_y - tl_y,
            RuntimeError(f"patch height({patch_height}) should be bigger than height({br_y - tl_y})"),
            parent_fn=get_centric_patches.__name__,
        )

    centric_patches_coord = []
    centric_patches_num_data = 0
    labels_in_patches = []

    # for points_dict in points:
    #     _points = points_dict["points"]

    #     [cx, cy] = get_center_point_from_points(_points)

    #     if len(_points) == 1:
    #         x, y = _points[0]
    #         if x < tl_x:
    #             x = tl_x
    #         elif x > br_x:
    #             x = br_x

    #         if y < tl_y:
    #             y = tl_y
    #         elif y > br_y:
    #             y = br_y
    #         cx, cy = x, y

    #     # if roi is not None and not is_point_in_roi([cx, cy], roi): # just for in case,
    #     #     return [], 0, []
    #     if logger is not None:
    #         logger.assertion_log(
    #             is_point_in_roi([cx, cy], [tl_x, tl_y, br_x, br_y]),
    #             RuntimeError(f"Center point({cx, cy}) of points must not be out of RoI({roi})"),
    #             parent_fn=get_centric_patches.__name__,
    #         )

    #     patch_coord = finetune_patch_for_roi(cx, cy, patch_width, patch_height, br_x, br_y, logger)
    #     points_in_patch = get_points_in_patch(
    #         points,
    #         patch_coord,
    #         [tl_x, tl_y, br_x, br_y],
    #         include_point_positive=include_point_positive,
    #         filename=filename,
    #     )
    #     if len(points_in_patch) != 0:
    #         labels_in_patches.append(points_in_patch)
    #         centric_patches_coord.append(patch_coord)
    #         centric_patches_num_data += 1

    # return centric_patches_coord, centric_patches_num_data, labels_in_patches

    for points_dict in points:
        _points = points_dict["points"]

        _xs, _ys = [], []
        for _point in _points:
            _xs.append(_point[0])
            _ys.append(_point[1])

        width = np.max(_xs) - np.min(_xs)
        height = np.max(_ys) - np.min(_ys)

        if width >= patch_width or height >= patch_height:
            roi_margin_x, roi_margin_y = 50, 50
            roi_x1 = int(min(tl_x, np.min(_xs)))
            roi_y1 = int(min(tl_y, np.min(_ys)))
            if roi_x1 - roi_margin_x > tl_x:
                roi_x1 -= roi_margin_x
            else:
                roi_x1 -= roi_margin_x - abs(roi_x1 - roi_margin_x - tl_x)

            if roi_y1 - roi_margin_y > tl_x:
                roi_y1 -= roi_margin_y
            else:
                roi_y1 -= roi_margin_y - abs(roi_y1 - roi_margin_y - tl_y)

            roi_x2 = int(max(br_x, np.max(_xs)))
            roi_y2 = int(max(br_y, np.max(_ys)))
            if roi_x2 + roi_margin_x < br_x:
                roi_x2 += roi_margin_x
            else:
                roi_x2 += roi_margin_x - abs(roi_x2 + roi_margin_x - br_x)

            if roi_y2 + roi_margin_y < br_y:
                roi_y2 += roi_margin_y
            else:
                roi_y2 += roi_margin_y - abs(roi_y2 + roi_margin_y - br_y)

            patch_coord, patches_num_data, points_in_patch = get_sliding_patches_v2(
                0,
                0,
                patch_height,
                patch_width,
                # [points_dict],
                points,
                roi=[roi_x1, roi_y1, roi_x2, roi_y2],
            )

            if len(points_in_patch) != 0:
                labels_in_patches += points_in_patch
                centric_patches_coord += patch_coord
                centric_patches_num_data += patches_num_data

        else:
            [cx, cy] = get_center_point_from_points(_points)

            if len(_points) == 1:
                x, y = _points[0]
                if x < tl_x:
                    x = tl_x
                elif x > br_x:
                    x = br_x

                if y < tl_y:
                    y = tl_y
                elif y > br_y:
                    y = br_y
                cx, cy = x, y

            if logger is not None:
                logger.assertion_log(
                    is_point_in_roi([cx, cy], [tl_x, tl_y, br_x, br_y]),
                    RuntimeError(f"Center point({cx, cy}) of points must not be out of RoI({roi})"),
                    parent_fn=get_centric_patches.__name__,
                )

            patch_coord = finetune_patch_for_roi(cx, cy, patch_width, patch_height, br_x, br_y, logger)
            points_in_patch = get_points_in_patch(
                points,
                patch_coord,
                [tl_x, tl_y, br_x, br_y],
                include_point_positive=include_point_positive,
                filename=filename,
            )
            if len(points_in_patch) != 0:
                labels_in_patches.append(points_in_patch)
                centric_patches_coord.append(patch_coord)
                centric_patches_num_data += 1

    return centric_patches_coord, centric_patches_num_data, labels_in_patches



import math

import torch
import torch.distributed as dist

from visionsuite.engines.classification.utils.registry import SAMPLERS


@SAMPLERS.register()
class RASampler(torch.utils.data.Sampler):
    """Sampler that restricts data loading to a subset of the dataset for distributed,
    with repeated augmentation.
    It ensures that different each augmented version of a sample will be visible to a
    different process (GPU).
    Heavily based on 'torch.utils.data.DistributedSampler'.

    This is borrowed from the DeiT Repo:
    https://github.com/facebookresearch/deit/blob/main/samplers.py
    """

    def __init__(self, dataset, num_replicas=None, rank=None, shuffle=True, seed=0, repetitions=3):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available!")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available!")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.num_samples = int(math.ceil(len(self.dataset) * float(repetitions) / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas
        self.num_selected_samples = int(math.floor(len(self.dataset) // 256 * 256 / self.num_replicas))
        self.shuffle = shuffle
        self.seed = seed
        self.repetitions = repetitions

    def __iter__(self):
        if self.shuffle:
            # Deterministically shuffle based on epoch
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        # Add extra samples to make it evenly divisible
        indices = [ele for ele in indices for i in range(self.repetitions)]
        indices += indices[: (self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # Subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices[: self.num_selected_samples])

    def __len__(self):
        return self.num_selected_samples

    def set_epoch(self, epoch):
        self.epoch = epoch



from .default import get_default_callbacks

class Callbacks:

    @classmethod
    def add_integration_callbacks(cls, callbacks_list):
        for callbacks in callbacks_list:
            for k, v in callbacks.items():
                if v not in cls._callbacks[k]:
                    cls._callbacks[k].append(v)

    def __init__(self, _callbacks=None, logger=None, logging_config=None):
        if logger is not None:
            self.logger = logger
        else:
            try:
                self.set_logger(
                    log_dir=logging_config.logs_dir,
                    log_stream_level=logging_config.log_stream_level,
                    log_file_level=logging_config.log_file_level,
                )
            except Exception as e:
                print(f"There has been error when creating logger in callbacks: {e}")
                self.logger = None

        self._callbacks = get_default_callbacks()

        if _callbacks is not None:
            self.add_callbacks(_callbacks)

    def add_callback(self, event: str, callback):

        assert (
            event in self._callbacks
        ), f"event '{event}' not found in callbacks {self._callbacks}"
        
        if isinstance(callback, list):
            for _callback in callback:
                assert callable(_callback), RuntimeError(f"{event} must be callable")
                self._callbacks[event].append(_callback)
        else:
            assert callable(callback), RuntimeError(f"{event} must be callable")
            self._callbacks[event].append(callback)

    def add_callbacks(self, callbacks):
        for event, callback in callbacks.items():
            self.add_callback(event, callback)

    def set_callback(self, event: str, callback):

        assert (
            event in self._callbacks
        ), f"event '{event}' not found in callbacks {self._callbacks}"
        assert callable(callback), RuntimeError(f"{event} must be callable")

        self._callbacks[event] = callback

    def run_callbacks(self, event: str, *args, **kwargs):
        for callback in self._callbacks.get(event, []):
            callback(self, *args, **kwargs)


#!/usr/bin/env python3

from .compose import Compose
from .formatting import (
    EdgeFormatBundle,
    BinaryEdgeFormatBundle,
    FormatEdge,
    FormatImage,
    FormatBinaryEdge,
    JointFormatBundle,
    JointBinaryFormatBundle,
)
from .loading import LoadAnnotations
from .transforms import (
    AddIgnoreBorder,
    Pad,
    RandomRotate,
    Resize,
)

__all__ = [
    "Compose",
    "EdgeFormatBundle",
    "BinaryEdgeFormatBundle",
    "FormatEdge",
    "FormatImage",
    "FormatBinaryEdge",
    "JointFormatBundle",
    "JointBinaryFormatBundle",
    "LoadAnnotations",
    "AddIgnoreBorder",
    "Pad",
    "RandomRotate",
    "Resize",
]



########################################################
#                                                      #
#       author: omitted for anonymous submission       #
#                                                      #
#     credits and copyright coming upon publication    #
#                                                      #
########################################################

import torch
from torch import nn

from src.models.model_one_modality import OWSNetwork
from src.models.resnet import ResNet

# from torchsummary import summary
from torchinfo import summary


def build_model(args, n_classes, verbose=True):
    if not args.pretrained_on_imagenet or args.last_ckpt:
        pretrained_on_imagenet = False
    else:
        pretrained_on_imagenet = True

    # set the number of channels in the encoder and for the
    # fused encoder features
    if "decreasing" in args.decoder_channels_mode:
        if args.decoder_channels_mode == "decreasing":
            channels_decoder = [512, 256, 128]

        print(
            "Notice: argument --channels_decoder is ignored when "
            "--decoder_chanels_mode decreasing is set."
        )
    else:
        channels_decoder = [args.channels_decoder] * 3

    if isinstance(args.nr_decoder_blocks, int):
        nr_decoder_blocks = [args.nr_decoder_blocks] * 3
    elif len(args.nr_decoder_blocks) == 1:
        nr_decoder_blocks = args.nr_decoder_blocks * 3
    else:
        nr_decoder_blocks = args.nr_decoder_blocks
        assert len(nr_decoder_blocks) == 3

    input_channels = 3
    model = OWSNetwork(
        height=args.height,
        width=args.width,
        pretrained_on_imagenet=pretrained_on_imagenet,
        encoder=args.encoder,
        encoder_block=args.encoder_block,
        activation=args.activation,
        input_channels=input_channels,
        encoder_decoder_fusion=args.encoder_decoder_fusion,
        context_module=args.context_module,
        num_classes=n_classes,
        pretrained_dir=args.pretrained_dir,
        nr_decoder_blocks=nr_decoder_blocks,
        channels_decoder=channels_decoder,
        upsampling=args.upsampling,
    )

    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    print("Device:", device)
    print("\n\n")

    model.to(device)
    # print(model)
    # summary(model)
    print("Model:", summary(model, (1, 3, args.height, args.width)))
    print("Number of parameters:", summary(model, (1, 3, args.height, args.width)).total_params)
    print("\n\n")
    if args.he_init:
        module_list = []

        # first filter out the already pretrained encoder(s)
        for c in model.children():
            if pretrained_on_imagenet and isinstance(c, ResNet):
                # already initialized
                continue
            for m in c.modules():
                module_list.append(m)

        # iterate over all the other modules
        # output layers, layers followed by sigmoid (in SE block) and
        # depthwise convolutions (currently only used in learned upsampling)
        # are not initialized with He method
        for i, m in enumerate(module_list):
            if isinstance(m, (nn.Conv2d, nn.Conv1d, nn.Linear)):
                if (
                    m.out_channels == n_classes
                    or m.out_channels == 19
                    or isinstance(module_list[i + 1], nn.Sigmoid)
                    or m.groups == m.in_channels
                ):
                    continue
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        print("Applied He init.")

    if args.finetune is not None:
        checkpoint = torch.load(args.finetune)
        model.load_state_dict(checkpoint["state_dict"])
        print(f"Loaded weights for finetuning: {args.finetune}")

        print("Freeze the encoder(s).")
        for name, param in model.named_parameters():
            if "encoder_rgb" in name:
                param.requires_grad = False

    return model, device



import torch
import torchvision 
import os.path as osp

from visionsuite.engines.segmentation.utils.registry import MODELS
from visionsuite.engines.utils.helpers import assert_key_dict
from visionsuite.engines.utils.bases.base_oop_module import BaseOOPModule

        
@MODELS.register()
class TorchvisionModel(BaseOOPModule):
    """
    - weights can be referred by https://pytorch.org/vision/main/models.html
    """
    def __init__(self):
        super().__init__(name='TorchvisionModel')
        self.args = None
        self._device = 'cpu'
        self._model_without_ddp = None
        self._model_ema = None
        self._model = None 
        self._params_to_optimize = None
               
    def build(self, *args, **kwargs):
        super().build(*args, **kwargs)
        
        
        self._load_model() 
        self._to_device()
        self._set_dist()
        self._set_params_to_optimize()
        
        self._load_seed_model()
            
    @property
    def model(self):
        return self._model
    
    @property
    def model_without_ddp(self):
        return self._model_without_ddp
    
    @property
    def model_ema(self):
        return self._model_ema
    
    @property
    def params_to_optimize(self):
        return self._params_to_optimize
    
    @BaseOOPModule.track_status
    def _load_model(self):
        try:
            self._model = torchvision.models.get_model(
                                    self.args['model_name'],
                                    weights=self.args['weights'],
                                    weights_backbone=self.args['weights_backbone'],
                                    num_classes=self.args['num_classes'],
                                    aux_loss=self.args['aux_loss'],
                )
            
            self.log_info(f"LOADED model: {self.args['model_name']}", self._load_model.__name__, __class__.__name__)
        except Exception as error:
            raise RuntimeError(f"{error}: There has been error when loading torchvision model: {self.args['type']} with config({self.args}): ")
           
    @BaseOOPModule.track_status
    def _to_device(self):
        assert_key_dict(self.args['train'], 'device')
        self._device = self.args['train']['device']
        self._model.to(self._device)
        
        self.log_info(f"TRANSFER model to device: {self._device}", self._to_device.__name__, __class__.__name__)

    @BaseOOPModule.track_status
    def _set_dist(self):
        assert_key_dict(self.args, 'distributed')
        assert_key_dict(self.args['train'], 'sync_bn')

        if self.args['distributed'] and self.args['train']['sync_bn']:
            self._model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self._model)
            self.log_info(f"CONVERT batchnorm into sync_batchnorm: {self._device}", self._set_dist.__name__, __class__.__name__)
            
        self._model_without_ddp = self._model
        if self.args['distributed']:
            assert_key_dict(self.args['distributed'], 'gpu')
            self._model = torch.nn.parallel.DistributedDataParallel(self._model, device_ids=[self.args['distributed']['gpu']])
            self.log_info(f"SET distributed Data Parallel: {self.args['distributed']['gpu']}", self._set_dist.__name__, __class__.__name__)
            self._model_without_ddp = self.model.module
        else:
            assert_key_dict(self.args['train'], 'device_ids')
            if len(self.args['train']['device_ids']) > 1:
                self._model = torch.nn.parallel.DataParallel(self._model, device_ids=self.args['train']['device_ids'])
                self._model_without_ddp = self.model.module
                self.log_info(f"SET Data Parallel: {self.args['train']['device_ids']}", self._set_dist.__name__, __class__.__name__)
                
    def _set_params_to_optimize(self):
        self._params_to_optimize = [
            {"params": [p for p in self._model_without_ddp.backbone.parameters() if p.requires_grad]},
            {"params": [p for p in self._model_without_ddp.classifier.parameters() if p.requires_grad]},
        ]
        self.log_info(f"SET params to optimize", self._set_params_to_optimize.__name__, __class__.__name__)
        
    def apply_aux_loss_to_params_to_optimize(self, lr):
        if self.args['aux_loss']:
            params = [p for p in self._model_without_ddp.aux_classifier.parameters() if p.requires_grad]
            self._params_to_optimize.append({"params": params, "lr": lr * 10})
            self.log_info(f"APPLY aux_loss to params to optimize: {self.args['aux_loss']}", self.apply_aux_loss_to_params_to_optimize.__name__, __class__.__name__)
            
    def _load_seed_model(self):
        if 'seed_model' in self.args and osp.exists(self.args['seed_model']):
            from visionsuite.engines.utils.torch_utils.utils import load_ckpt

            assert osp.exists(self.args['seed_model']), ValueError(f"There is no such checkpoint: {self.args['seed_model']}")
            ckpt = load_ckpt(self.args['seed_model'])
            self.log_info(f"LOADED ckpt: {ckpt.keys()}", self._load_seed_model.__name__, __class__.__name__)
            
            self._model_without_ddp.load_state_dict(ckpt['model'], strict=True)
            self._model.load_state_dict(ckpt['model'], strict=True)
            self.log_info(f"LOADED seed_model weights", self._load_seed_model.__name__, __class__.__name__)
                    
        else:
            self.log_info(f"NOT LOADED seed_model", self._load_seed_model.__name__, __class__.__name__)   
    


#!/usr/bin/env python3

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule
from mmseg.models.decode_heads.aspp_head import ASPPHead
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule
from mmseg.ops import resize

from ..builder import HEADS


@HEADS.register_module()
class ModDepthwiseSeparableASPPHead(ASPPHead):
    """Modded DeepLabV3+ Head

    Encoder-Decoder with Atrous Separable Convolution for Semantic Image
    Segmentation.
    This head is the implementation of `DeepLabV3+
    <https://arxiv.org/abs/1802.02611>`_.
    Args:
        c1_in_channels (int): The input channels of c1 decoder. If is 0,
            the no decoder will be used.
        c1_channels (int): The intermediate channels of c1 decoder.
        c1_in_index (int): index of c1 feature
    """

    def __init__(self, c1_in_channels, c1_channels, c1_in_index=0, **kwargs):
        super(ModDepthwiseSeparableASPPHead, self).__init__(**kwargs)
        assert c1_in_channels >= 0

        self.c1_in_index = c1_in_index

        self.aspp_modules = DepthwiseSeparableASPPModule(
            dilations=self.dilations,
            in_channels=self.in_channels,
            channels=self.channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        if c1_in_channels > 0:
            self.c1_bottleneck = ConvModule(
                c1_in_channels,
                c1_channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            )
        else:
            self.c1_bottleneck = None
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                self.channels + c1_channels,
                self.channels,
                3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                self.channels,
                self.channels,
                3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

    def forward(self, inputs):
        """Forward function."""
        x = self._transform_inputs(inputs)
        aspp_outs = [
            resize(
                self.image_pool(x),
                size=x.size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.extend(self.aspp_modules(x))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck(aspp_outs)
        if self.c1_bottleneck is not None:
            c1_output = self.c1_bottleneck(inputs[self.c1_in_index])
            output = resize(
                input=output,
                size=c1_output.shape[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
            output = torch.cat([output, c1_output], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)
        return output
